from abc import ABC
from typing import Union, Text, MutableMapping, Any, Iterable, Tuple, Optional, IO, Callable, List, TypeVar, Type

from requests import Request, Response, PreparedRequest
from requests import Session as OldSession
from requests import auth as _auth
from requests.cookies import RequestsCookieJar

DataType = TypeVar("DataType")

class SessionHook(ABC):
    pass


class RequestHook(SessionHook):
    def run(self, request: Request) -> Request:
        ...


class PreparedRequestHook(SessionHook):
    def run(self, request: PreparedRequest) -> PreparedRequest:
        ...


class ResponseHook(SessionHook):
    def run(self, response: Response) -> Response:
        ...


class ApitistResponse(Response):
    data: DataType
    def __init__(self, response: Response): ...
    def verify_response(self, ok_status: Union[int, List[int]] = 200) -> "ApitistResponse": ...
    def vr(self, ok_status: Union[int, List[int]] = 200) -> "ApitistResponse": ...
    def structure(self, t: Type[DataType]) -> "ApitistResponse": ...

_Data = Union[None, Text, bytes, MutableMapping[str, Any], MutableMapping[Text, Any], Iterable[Tuple[Text, Optional[Text]]], IO]

_Hook = Callable[[ApitistResponse], Any]
_Hooks = MutableMapping[Text, List[_Hook]]
_HooksInput = MutableMapping[Text, Union[Iterable[_Hook], _Hook]]


class Session(OldSession):
    request_hooks: List[RequestHook]
    prep_request_hooks: List[PreparedRequestHook]
    response_hooks: List[ResponseHook]
    def add_request_hook(self, hook: Type[RequestHook]): ...
    def add_prep_request_hook(self, hook: Type[PreparedRequestHook]): ...
    def add_response_hook(self, hook: Type[ResponseHook]): ...
    def add_hook(self, hook: Type[Union[RequestHook, PreparedRequestHook, ResponseHook]]): ...
    def request(self, method: str, url: Union[str, bytes, Text],
                params: Union[None, bytes, MutableMapping[Text, Text]] = ...,
                data: _Data = ...,
                headers: Optional[MutableMapping[Text, Text]] = ...,
                cookies: Union[None, RequestsCookieJar, MutableMapping[Text, Text]] = ...,
                files: Optional[MutableMapping[Text, IO[Any]]] = ...,
                auth: Union[None, Tuple[Text, Text], _auth.AuthBase, Callable[[Request], Request]] = ...,
                timeout: Union[None, float, Tuple[float, float], Tuple[float, None]] = ...,
                allow_redirects: Optional[bool] = ...,
                proxies: Optional[MutableMapping[Text, Text]] = ...,
                hooks: Optional[_HooksInput] = ...,
                stream: Optional[bool] = ...,
                verify: Union[None, bool, Text] = ...,
                cert: Union[Text, Tuple[Text, Text], None] = ...,
                json: Optional[Any] = ...,
                ) -> ApitistResponse: ...
    def get(self, url: Union[Text, bytes], **kwargs) -> ApitistResponse: ...
    def options(self, url: Union[Text, bytes], **kwargs) -> ApitistResponse: ...
    def head(self, url: Union[Text, bytes], **kwargs) -> ApitistResponse: ...
    def post(self, url: Union[Text, bytes], data: _Data = ..., json: Optional[Any] = ..., **kwargs) -> ApitistResponse: ...
    def put(self, url: Union[Text, bytes], data: _Data = ..., **kwargs) -> ApitistResponse: ...
    def patch(self, url: Union[Text, bytes], data: _Data = ..., **kwargs) -> ApitistResponse: ...
    def delete(self, url: Union[Text, bytes], **kwargs) -> ApitistResponse: ...


def session() -> Session: ...



import logging

import apitist


class Logging:

    LOG_LEVEL = logging.INFO
    logger = None

    @staticmethod
    def set_logging_level(value):
        Logging.LOG_LEVEL = value
        Logging._setup_logging(Logging.LOG_LEVEL)

    @staticmethod
    def _setup_logging(log_level):
        """Setup basic logging

        Args:
          log_level (int): minimum loglevel for emitting messages
        """

        # Suppress overly verbose logs from libraries that aren't helpful
        if log_level == logging.DEBUG:
            logging.getLogger("requests").setLevel(logging.DEBUG)
            logging.getLogger("urllib3").setLevel(logging.DEBUG)
        else:
            logging.getLogger("requests").setLevel(logging.WARNING)
            logging.getLogger("urllib3").setLevel(logging.WARNING)

        Logging.logger = logging.getLogger(apitist.dist_name)
        Logging.logger.setLevel(log_level)


Logging._setup_logging(Logging.LOG_LEVEL)



import inspect
import random as rnd
import typing
from dataclasses import MISSING, fields, is_dataclass

import attr

from apitist.utils import (
    has_args,
    is_attrs_class,
    is_sequence,
    is_tuple,
    is_union_type,
)

from .logging import Logging

T = typing.TypeVar("T")


class Randomer:
    _types_dict = None

    def __init__(self):
        self._types_dict = dict()

    @property
    def available_hooks(self):
        return self._types_dict

    def get_hook(self, t: typing.Type):
        """Get function for given type"""
        func = self._types_dict.get(t)
        if func is None:
            raise TypeError("Unable to find hook for {} type".format(t))
        return func

    def run_hook(self, t: typing.Type, **set_params):
        """Generate random data for given type"""
        func = self.get_hook(t)
        if inspect.getfullargspec(func).varkw:
            Logging.logger.debug(
                "All additional params will be passed to hook: %s", set_params
            )
            res = func(**set_params)
        else:
            res = func()
        Logging.logger.debug("Generated data for type %s: %s", t, res)
        return res

    def add_type(self, t: typing.Type[T], func: typing.Callable[[], T]):
        """
        Add new type for random generation.
        Function should return given type.
        """
        if t in self._types_dict:
            Logging.logger.warning(
                "Type %s already exists in dict, overriding", t
            )
        Logging.logger.debug("Registering type %s with function %s", t, func)
        self._types_dict[t] = func

    def add_types(
        self, types_dict: typing.Dict[typing.Type[T], typing.Callable[[], T]]
    ):
        """
        Add new types for random generation.
        Functions should return given type.
        """
        Logging.logger.debug("Registering list of types: %s", types_dict)
        self._types_dict.update(types_dict)

    def random_object(
        self,
        t: typing.Type[T],
        required_only=False,
        ignore: typing.List[str] = None,
        only: typing.List[str] = None,
        inverse=False,
        **set_params,
    ) -> T:
        """
        Create object of given type with random data

        Be careful, random object does not use converter to create a type
        It may lead to types missmatch

        :param t: Type which would me generated randomly
        :param required_only: Use only fields which do not have default values
        :param ignore: List of fields which should be ignored
            (will be passed to hook kwargs)
        :param only: List of fields which should be used
            (just a workaround near ignore + inverse)
        :param inverse: Inverse ignore list (will be passed to hook kwargs)
        :param set_params: Custom params which would be manually set to type
            (will be passed to hook kwargs)
        :return: Object of given type
        """
        Logging.logger.debug("Generating random data for type %s", t)
        if only and ignore:
            raise ValueError(
                "Only one of parameters - only or ignore should be used"
            )
        if only:
            ignore = only
            inverse = True
        if ignore is None:
            ignore = list()
        if t in self.available_hooks:
            return self.run_hook(
                t, ignore=ignore, inverse=inverse, **set_params
            )
        elif is_attrs_class(t):
            data = {}
            for field in attr.fields(t):
                f_name = field.name
                has_default = field.default is not attr.NOTHING
                if f_name in set_params:
                    data[f_name] = set_params[f_name]
                    continue
                if (
                    (f_name in ignore and inverse is False)
                    or (ignore and f_name not in ignore and inverse is True)
                    or (required_only and has_default)
                ):
                    data[f_name] = field.default if has_default else None
                    continue
                data[f_name] = self.random_object(
                    field.type,
                    required_only=required_only,
                    ignore=ignore,
                    inverse=inverse,
                    **set_params,
                )
            data = t(**data)
            Logging.logger.debug(
                "Generating random data for attrs type %s", data
            )
            return data
        elif is_dataclass(t):
            data = {}
            for field in fields(t):
                f_name = field.name
                has_default = (
                    field.default is not MISSING
                    or field.default_factory is not MISSING
                )
                if f_name in set_params:
                    data[f_name] = set_params[f_name]
                    continue
                if (
                    (f_name in ignore and inverse is False)
                    or (ignore and f_name not in ignore and inverse is True)
                    or (required_only and has_default)
                ):
                    data[f_name] = None
                    if has_default:
                        data[f_name] = (
                            field.default
                            if field.default != MISSING
                            else field.default_factory()
                        )
                    continue
                data[f_name] = self.random_object(
                    field.type,
                    required_only=required_only,
                    ignore=ignore,
                    inverse=inverse,
                    **set_params,
                )
            data = t(**data)
            Logging.logger.debug(
                "Generating random data for dataclass type %s", data
            )
            return data
        elif is_union_type(t) and has_args(t):
            return self.random_object(
                rnd.choice(t.__args__),
                required_only=required_only,
                ignore=ignore,
                inverse=inverse,
                **set_params,
            )
        elif is_tuple(t) and has_args(t):
            return (
                self.random_object(
                    t.__args__[0],
                    required_only=required_only,
                    ignore=ignore,
                    inverse=inverse,
                    **set_params,
                ),
            )
        elif is_sequence(t) and has_args(t):
            return [
                self.random_object(
                    t.__args__[0],
                    required_only=required_only,
                    ignore=ignore,
                    inverse=inverse,
                    **set_params,
                )
            ]
        return None

    object = random_object

    def random_partial(
        self, t: typing.Type[T], use: list = (), **set_params
    ) -> T:
        """
        Create object of given type with random data with only given fields.
        """
        if t in self.available_hooks:
            return self.run_hook(t, use=use, **set_params)
        elif is_attrs_class(t):
            data = {}
            for field in attr.fields(t):
                key = field.name
                if set_params and key in set_params:
                    data[key] = set_params[key]
                    continue
                if (
                    use
                    and key not in use
                    and (
                        not is_attrs_class(field.type)
                        or (
                            is_attrs_class(field.type)
                            and field.type in self.available_hooks
                        )
                    )
                ):
                    data[key] = attr.NOTHING
                    continue
                data[key] = self.random_partial(field.type, use, **set_params)

            if not all(v == attr.NOTHING for _, v in data.items()):
                return t(**data)
            return attr.NOTHING

        elif is_dataclass(t):
            data = {}
            for field in fields(t):
                key = field.name
                if set_params and key in set_params:
                    data[key] = set_params[key]
                    continue
                if (
                    use
                    and key not in use
                    and (
                        not is_dataclass(field.type)
                        or (
                            is_dataclass(field.type)
                            and field.type in self.available_hooks
                        )
                    )
                ):
                    data[key] = MISSING
                    continue

                data[key] = self.random_partial(field.type, use, **set_params)

            if not all(v == MISSING for _, v in data.items()):
                return t(**data)
            return MISSING

        elif is_union_type(t) and has_args(t):
            return self.random_partial(
                rnd.choice(t.__args__), use=use, **set_params
            )
        elif is_tuple(t) and has_args(t):
            return (self.random_partial(t.__args__[0], use=use, **set_params),)
        elif is_sequence(t) and has_args(t):
            return [self.random_partial(t.__args__[0], use=use, **set_params)]
        return None

    partial = random_partial
